import time

from .util import choose_population, decode_http_header, decode_url
from .record import PageRecord, PixelRecord, GoalRecord, SplitRecord


# Goal Value Aggregation Types
RATE = 'R'
AVERAGE = 'A'
SUM = 'S'
PER = 'I'

# Goal Value Measurement Formats
NUMERIC = 'N'
CURRENCY = 'C'
PERCENTAGE = 'P'


class Visitor(object):
    """
    A handle to perform operations on the given visitor session.
    """
    RATE = RATE
    AVERAGE = AVERAGE
    SUM = SUM
    PER = PER

    NUMERIC = NUMERIC
    CURRENCY = CURRENCY
    PERCENTAGE = PERCENTAGE

    def __init__(self, id, log, site_id='', buffer_writes=True):
        """
        Initialize the Visitor handle.

        :param id:
            id to reference this Visitor.
        :type id:
            str
        :param log:
            A log instance that implements the manhattan log interface methods.
        :param buffer_writes:
            If True, log entries are buffered and flushed at the end of
            a request or when `flush()` is called manually; if False,
            they're written immediately.
        """
        self.id = id
        self.log = log
        self.site_id = str(site_id)
        self.buffer_writes = buffer_writes
        self.buffer = []

    def timestamp(self):
        """
        Override this to generate event timestamps in a different way. Defaults
        to the POSIX epoch.
        """
        return '%0.4f' % time.time()

    def write(self, *records):
        self.buffer += records
        if not self.buffer_writes:
            self.flush()

    def flush(self):
        """Write buffered records to log."""
        if self.buffer:
            records = [r.to_list() for r in self.buffer]
            self.buffer = []
            self.log.write(*records)

    def page(self, request):
        """
        Log a page view for this visitor.

        :param request:
            A request object corresponding to the page to log.
        :type request:
            webob.Request instance
        """
        rec = PageRecord(timestamp=self.timestamp(),
                         vid=self.id,
                         site_id=self.site_id,
                         ip=request.remote_addr or '0.0.0.0',
                         method=request.method,
                         url=decode_url(request.url),
                         user_agent=decode_http_header(request.user_agent),
                         referer=decode_http_header(request.referer))
        self.write(rec)

    def pixel(self):
        """
        Log a pixel view for this visitor.
        """
        rec = PixelRecord(timestamp=self.timestamp(),
                          vid=self.id,
                          site_id=self.site_id)
        self.write(rec)

    def goal(self, name, value=None, value_type=None, value_format=None):
        """
        Log a goal hit for this visitor.

        :param name:
            Name of the goal.
        :type name:
            str
        :param value:
            Value of this goal.
        :type value:
            int or float
        :param value_type:
            Type of goal value aggregation to perform.
        :type value_type:
            RATE, AVERAGE or SUM
        :param value_format:
            Display format for this goal value.
        :type value_format:
            NUMERIC, CURRENCY, or PERCENTAGE
        """
        value = value and str(value)
        rec = GoalRecord(timestamp=self.timestamp(),
                         vid=self.id,
                         site_id=self.site_id,
                         name=name.encode('ascii', 'replace').decode('ascii'),
                         value=value or '',
                         value_type=value_type or '',
                         value_format=value_format or '')
        self.write(rec)

    def split(self, test_name, populations=None):
        """
        Perform a split test for this visitor. The resulting population is
        calculated deterministically based on the test name and the visitor id,
        so the same visitor id and the same test name will always be assigned
        to the same population.

        :param test_name:
            Name of the test.
        :type test_name:
            str
        :param populations:
            Population specified. Can be any of the following:

                None -- 50/50 split performed between True or False.
                list -- Select evenly between entries in the list.
                dict -- A weighted split between keys in the dict. The weight
                of each population is specified by the value, as a float.

        :returns:
            The population selected for the visitor.
        """
        selected = choose_population(self.id + test_name, populations)
        rec = SplitRecord(timestamp=self.timestamp(),
                          vid=self.id,
                          site_id=self.site_id,
                          test_name=test_name,
                          selected=str(selected))
        self.write(rec)
        return selected



from webob import Request, Response

from .visitor import Visitor
from .util import nonce, transparent_pixel, pixel_tag, Signer


class ManhattanMiddleware(object):

    def __init__(self, app, log, secret, cookie_name='manhattan',
                 pixel_path='/vpixel.gif', host_map=None, buffer_writes=True):
        self.app = app
        self.cookie_name = cookie_name
        self.log = log
        self.signer = Signer(secret)
        self.pixel_path = pixel_path
        self.host_map = host_map or {}
        self.buffer_writes = buffer_writes

    def inject_pixel(self, resp):
        tag = pixel_tag(self.pixel_path)

        def wrap_iter(orig_iter):
            for chunk in orig_iter:
                yield chunk.replace('</body>', '%s</body>' % tag)

        resp.app_iter = wrap_iter(resp.app_iter)

    def handle_pixel(self, visitor, fresh):
        if not fresh:
            visitor.pixel()
        resp = Response(transparent_pixel)
        resp.content_type = 'image/gif'
        return resp

    def count_page(self, req):
        return (req.method in ('GET', 'POST') and
                req.headers.get('X-Purpose') != 'preview')

    def get_visitor_id(self, req):
        signed_value = req.cookies[self.cookie_name]
        return self.signer.unsign(signed_value)

    def __call__(self, environ, start_response):
        req = Request(environ)

        if self.cookie_name in req.cookies:
            vid = self.get_visitor_id(req)
            fresh = False
        else:
            vid = nonce()
            fresh = True

        site_id = self.host_map.get(req.host.split(':', 1)[0], 0)

        req.environ['manhattan.visitor'] = visitor = Visitor(
            vid, self.log, site_id, self.buffer_writes)

        if self.pixel_path and req.path_info == self.pixel_path:
            resp = self.handle_pixel(visitor, fresh)
        else:
            resp = req.get_response(self.app)

            if self.count_page(req):
                visitor.page(req)

            if fresh:
                resp.set_cookie(self.cookie_name, self.signer.sign(visitor.id),
                                httponly=True)

            if self.pixel_path and resp.content_type == 'text/html':
                self.inject_pixel(resp)

        visitor.flush()
        return resp(environ, start_response)



import types
from random import randint

from webob import Request

from manhattan import visitor
from manhattan.visitor import Visitor

from .base import work_path


test_complex_goals = [
    (u'abandoned cart', set([u'add to cart']), set([u'began checkout'])),
    (u'abandoned checkout',
     set([u'began checkout']), set([u'completed checkout'])),
    (u'abandoned after validation failure',
     set([u'began checkout', u'checkout validation failed']),
     set([u'completed checkout'])),
    (u'abandoned after payment failure',
     set([u'began checkout', u'payment failed']),
     set([u'completed checkout'])),
]


sampleconfig = {
    'input_log_path': work_path('clientserver-python-config'),
    'sqlalchemy_url': 'sqlite:///' + work_path('sampleconfig.db'),
    'complex_goals': test_complex_goals,
    'bind': 'tcp://127.0.0.1:5556',
    'verbose': False,
    'error_log_path': work_path('python-config-debug.log'),
}


test_clickstream = [
    (1, 10, 'page', 'a', '/'),
    (1, 10, 'page', 'b', '/cheese'),
    (1, 132, 'pixel', 'a'),
    (1, 240, 'page', 'a', '/potatoes'),
    (1, 290, 'page', 'a', '/potatoes/sweet'),
    (1, 295, 'page', 'bot', '/potatoes/russet'),
    (1, 300, 'pixel', 'b'),
    (1, 382, 'page', 'a', '/potatoes/russet'),
    (1, 385, 'goal', 'a', 'add to cart', ''),
    (1, 394, 'page', 'b', '/cheese/parmesan'),
    (2, 401, 'page', 'q', '/'),
    (1, 448, 'page', 'c', '/fruit/apples'),
    (1, 462, 'page', 'b', '/cart'),
    (2, 544, 'page', 'q', '/candy'),
    (2, 545, 'pixel', 'q'),
    (2, 680, 'page', 'q', '/candy'),
    (1, 749, 'goal', 'f', 'fake goal', ''),
    (1, 1120, 'goal', 'b', 'add to cart', ''),
    (1, 1180, 'page', 'bot', '/potatoes/russet'),
    (1, 1200, 'page', 'f', '/'),
    (1, 1202, 'pixel', 'f'),
    (2, 1311, 'page', 'q', '/candy/molds'),
    (1, 1596, 'page', 'c', '/fruit/bananas'),
    (1, 1602, 'split', 'b', 'red checkout form'),
    (1, 1602, 'page', 'b', '/checkout'),
    (1, 1602, 'goal', 'b', 'began checkout', ''),
    (1, 1706, 'page', 'f', '/fruit'),
    (2, 1807, 'page', 'q', '/cart'),
    (2, 1807, 'goal', 'q', 'add to cart', ''),
    (1, 1821, 'page', 'bot', 'fruit'),
    (1, 1920, 'page', 'bot', '/cart'),
    (1, 1950, 'goal', 'bot', 'add to cart', ''),
    (1, 1996, 'page', 'a', '/cart'),
    (1, 1996, 'goal', 'a', 'add to cart', ''),
    (1, 2043, 'goal', 'b', 'checkout validation failed', ''),
    (1, 2112, 'pixel', 'c'),
    (1, 2196, 'page', 'a', '/cheese/gouda'),
    (1, 2356, 'page', 'a', '/cheese'),
    (2, 2477, 'page', 'q', '/checkout'),
    (2, 2477, 'goal', 'q', 'began checkout', ''),
    (1, 2680, 'page', 'b', '/cheese'),
    (1, 2840, 'page', 'd', '/'),
    (1, 2846, 'pixel', 'd'),
    (1, 3110, 'split', 'b', 'red checkout form'),
    (1, 3340, 'page', 'd', '/cheese'),
    (1, 3514, 'page', 'd', '/cheese/gruyere'),
    (1, 3514, 'page', 'b', '/checkout/complete'),
    (1, 3514, 'goal', 'b', 'completed checkout', 31.78),
    (1, 3514, 'goal', 'b', 'order margin', 22.5),
    (1, 3514, 'goal', 'b', 'margin per session', 7.15),
    (1, 3600, 'page', 'c', '/'),
    (1, 3620, 'page', 'd', '/cart'),
    (1, 4114, 'goal', 'd', 'add to cart', ''),
    (1, 4278, 'split', 'd', 'red checkout form'),
    (1, 4278, 'page', 'd', '/checkout'),
    (1, 4278, 'goal', 'd', 'began checkout', ''),
    (1, 4534, 'page', 'a', '/account'),
    (1, 4600, 'page', 'e', '/fruit'),
    (1, 4616, 'pixel', 'e'),
    (1, 4700, 'page', 'bot', '/fruit/cherries'),
    (1, 4990, 'split', 'd', 'red checkout form'),
    (1, 4990, 'page', 'd', '/checkout/complete'),
    (1, 4990, 'goal', 'd', 'completed checkout', 64.99),
    (1, 4990, 'goal', 'd', 'order margin', 20.1),
    (1, 4990, 'goal', 'd', 'margin per session', 13.06),
    (1, 5002, 'page', 'e', '/fruit/pineapples'),
    (1, 5174, 'page', 'e', '/fruit/cherries'),
    (1, 5226, 'page', 'e', '/fruit/pears'),
    (1, 5244, 'page', 'e', '/cart'),
    (1, 5244, 'goal', 'e', 'add to cart', ''),
    (1, 5950, 'split', 'e', 'red checkout form'),
    (1, 5950, 'page', 'e', '/checkout'),
    (1, 5950, 'goal', 'e', 'began checkout', ''),
    (1, 6278, 'page', 'd', '/account'),
    (1, 6396, 'page', 'd', '/'),
    (1, 6620, 'split', 'e', 'red checkout form'),
    (1, 6620, 'page', 'e', '/checkout/complete'),
    (1, 6620, 'goal', 'e', 'completed checkout', 11.42),
    (1, 6620, 'goal', 'e', 'order margin', 27.8),
    (1, 6620, 'goal', 'e', 'margin per session', 3.17),
    (1, 6988, 'page', 'b', '/fruit'),
    (1, 7020, 'page', 'f', '/cheese'),
    (1, 7042, 'page', 'f', '/cheese/cheddar'),
    (1, 7068, 'page', 'f', '/cart'),
    (1, 7068, 'goal', 'f', 'add to cart', ''),
    (1, 7198, 'page', 'f', '/cheese'),
    (1, 7246, 'split', 'f', 'red checkout form'),
    (1, 7246, 'page', 'f', '/checkout'),
    (1, 7246, 'goal', 'f', 'began checkout', ''),
    (1, 7350, 'goal', 'f', 'payment failed', ''),
]


def run_clickstream(log, first=None, last=None):
    value_types = {
        'completed checkout': visitor.SUM,
        'order margin': visitor.AVERAGE,
        'margin per session': visitor.PER,
    }

    value_formats = {
        'completed checkout': visitor.CURRENCY,
        'order margin': visitor.PERCENTAGE,
        'margin per session': visitor.CURRENCY,
    }

    browsers = {
        'a': u'Chrome/666.0',
        'b': u'Safari/12345',
        'c': u'Firefox/infinity',
        'd': u'Chrome/17',
        'e': u'Opera/sucks',
        'f': u'MSIE/9',
        'bot': u'ScroogleBot',
        'q': u'Chrome/641.1',
    }

    visitors = {}

    def get_visitor(vid, site_id):
        if vid not in visitors:
            visitors[vid] = Visitor(
                vid, log, site_id=site_id, buffer_writes=False)
        return visitors[vid]

    def set_fake_timestamp(v, ts):
        def fake_timestamp(self):
            return '%d.%04d' % (ts, randint(0, 9999))
        v.timestamp = types.MethodType(fake_timestamp, v)

    stream = test_clickstream
    if first and last:
        stream = stream[first:last]
    elif first:
        stream = stream[first:]
    elif last:
        stream = stream[:last]

    for action in stream:
        site_id = action[0]
        ts = action[1]
        cmd = action[2]
        vid = action[3]
        v = get_visitor(vid, site_id)
        args = action[4:]

        set_fake_timestamp(v, ts)

        if cmd == 'page':
            req = Request.blank(args[0])
            req.user_agent = browsers[vid]
            v.page(req)
        elif cmd == 'pixel':
            v.pixel()
        elif cmd == 'goal':
            goal_name = unicode(args[0])
            value = args[1]
            value_type = value_types.get(goal_name)
            value_format = value_formats.get(goal_name)
            v.goal(goal_name, value=value,
                   value_type=value_type,
                   value_format=value_format)
        elif cmd == 'split':
            v.split(unicode(args[0]))



import os.path
import shutil
from unittest import TestCase


work_dir = '/tmp/manhattan-tests'


def work_path(path):
    return os.path.join(work_dir, path)


class BaseTest(TestCase):

    def setUp(self):
        if os.path.exists(work_dir):
            shutil.rmtree(work_dir)



import math
from collections import defaultdict
from unittest import TestCase

from manhattan import util


class TestUtil(TestCase):

    def assertRandomish(self, s, bits=4):
        # Calculate entropy.
        counts = defaultdict(int)
        for char in s:
            counts[char] += 1

        n = float(len(s))
        entropy = sum((c / n) * math.log(n / c) / math.log(2)
                      for c in counts.values())
        self.assertLess(bits - entropy, .01)

    def assertRoughly(self, a, b, f=0.5):
        bot = b - (b * f)
        top = b + (b * f)
        self.assertLessEqual(bot, a)
        self.assertLessEqual(a, top)

    def test_nonce(self):
        n1 = util.nonce()
        n2 = util.nonce()
        self.assertNotEqual(n1, n2)

        s = b''.join(util.nonce() for i in range(100))
        self.assertRandomish(s)

    def test_choose_population_bool(self):
        a = 0
        for ii in range(200):
            if util.choose_population(util.nonce()):
                a += 1
        # Make sure it's relatively uniform...
        self.assertRoughly(a, 100)

    def test_chose_population_bad_value(self):
        with self.assertRaises(ValueError):
            util.choose_population(util.nonce(), 123)

    def test_choose_population_zero_mass(self):
        with self.assertRaises(ValueError):
            util.choose_population(util.nonce(), {'foo': 0})

    def test_choose_population_list(self):
        counts = defaultdict(int)
        for ii in range(300):
            choice = util.choose_population(util.nonce(),
                                            ['foo', 'bar', 'baz'])
            counts[choice] += 1
        self.assertRoughly(counts['foo'], 100)
        self.assertRoughly(counts['bar'], 100)
        self.assertRoughly(counts['baz'], 100)

    def test_choose_population_weighted(self):
        counts = defaultdict(int)
        for ii in range(300):
            choice = util.choose_population(util.nonce(), {'foo': 0.1,
                                                           'quux': 0,
                                                           'bar': 0.1,
                                                           'baz': 0.8})
            counts[choice] += 1
        self.assertRoughly(counts['foo'], 30)
        self.assertRoughly(counts['bar'], 30)
        self.assertRoughly(counts['baz'], 240)
        self.assertEqual(counts['quux'], 0)

    def test_decode_http_header_none(self):
        self.assertEqual(util.decode_http_header(None), u'')

    def test_decode_http_header(self):
        self.assertEqual(util.decode_http_header('hello \xf6 \xe1 world'),
                         u'hello \xf6 \xe1 world')


class TestSigner(TestCase):
    def setUp(self):
        self.sample = util.nonce()

    def test_round_trip(self):
        signer = util.Signer('s3krit')
        signed = signer.sign(self.sample)

        b = signer.unsign(signed)
        self.assertEqual(self.sample, b)

    def test_bad_signature(self):
        signer = util.Signer('s3krit')
        signed = signer.sign(self.sample)

        mangled = signed[:-3]
        with self.assertRaises(util.BadSignature) as cm:
            signer.unsign(mangled)

        self.assertIn(mangled, str(cm.exception))

    def test_lowercase(self):
        signer = util.Signer('s3krit')
        signed = signer.sign(self.sample)

        b = signer.unsign(signed.lower())
        self.assertEqual(self.sample, b)

    def test_uppercase(self):
        signer = util.Signer('s3krit')
        signed = signer.sign(self.sample)

        b = signer.unsign(signed.upper())
        self.assertEqual(self.sample, b)

    def test_bad_data(self):
        signer = util.Signer('s3krit')
        signed = signer.sign(self.sample)

        mangled = signed.split('.')[0]
        with self.assertRaises(util.BadData) as cm:
            signer.unsign(mangled)

        self.assertIn('No separator', str(cm.exception))
        self.assertIn(mangled, str(cm.exception))



log_version = 1


class Record(object):
    base_fields = ('timestamp', 'vid', 'site_id')
    fields = ()

    def __init__(self, **kwargs):
        for field in self.base_fields + self.fields:
            setattr(self, field, kwargs.get(field, ''))

    def to_list(self):
        return ([str(log_version), self.key] +
                [getattr(self, field) for field in
                 self.base_fields + self.fields])

    @staticmethod
    def from_list(vals):
        version = vals[0]
        record_type = vals[1]
        rest = vals[2:]

        assert int(version) == log_version

        cls = _record_types[record_type]
        kwargs = {field: val for field, val
                  in zip(cls.base_fields + cls.fields, rest)}
        return cls(**kwargs)


class PageRecord(Record):
    key = 'page'
    fields = ('ip', 'method', 'url', 'user_agent', 'referer')


class PixelRecord(Record):
    key = 'pixel'
    fields = ()


class GoalRecord(Record):
    key = 'goal'
    fields = ('name', 'value', 'value_type', 'value_format')


class SplitRecord(Record):
    key = 'split'
    fields = ('test_name', 'selected')


_record_types = {cls.key: cls for cls in
                 (PageRecord, PixelRecord, GoalRecord, SplitRecord)}



class VisitorHistory(object):
    """
    ``VisitorHistory`` objects store fully or partially aggregated state about
    the previous events related to a given visitor. Instances of this class are
    intended to be stored with a reference to the visitor, like in a key/value
    store keyed by vistor ID.
    """
    def __init__(self):
        self.nonbot = False
        self.nonbot_queue = []
        self.goals = set()
        self.variants = set()
        self.ips = set()
        self.user_agents = set()

        # Set of (goal name, rollup key, bucket start) that have already been
        # counted.
        self.conversion_keys = set()

        # Set of (test name, selected, rollup key, bucket start) that have
        # already been counted.
        self.impression_keys = set()

        # Set of (test name, selected, goal name, rollup key, bucket start)
        # that have already been counted.
        self.variant_conversion_keys = set()

        # Dict mapping complex goal name to a list of previous conversion keys
        # on that complex goal.
        self.complex_keys = {}


class Test(object):
    """
    Stores denormalized information about a particular test. Instances of this
    class area intended to be stored with a reference to the test name, like in
    a key/value store keyed by test name.
    """
    def __init__(self, first_timestamp=None, last_timestamp=None,
                 variants=None):
        self.first_timestamp = first_timestamp
        self.last_timestamp = last_timestamp
        self.variants = variants or set()


class Goal(object):
    """
    Stores denormalized information about a particular goal. Instances of this
    class area intended to be stored with a reference to the goal name, like in
    a key/value store keyed by goal name.
    """
    def __init__(self, value_type=None, value_format=None):
        self.value_type = value_type
        self.value_format = value_format



import code
import os
import os.path
import sys
import time

from threading import Event, Thread

from sqlalchemy import create_engine

from manhattan.server import Server, main as server_main, logging_config
from manhattan.client import (ServerError, TimeoutError, Client,
                              main as client_main)
from manhattan.log.timerotating import TimeRotatingLog

from . import data
from .base import BaseTest, work_path
from .test_combinations import drop_existing_tables


class MockBackend(object):

    def foo(self, a, b):
        return u"foo: %r %r" % (a, b)

    def bar(self, *args, **kw):
        return u"bar: %r %r" % (args, kw)

    def failme(self, a):
        raise ValueError('sad')


class TestClientServer(BaseTest):

    def test_basic(self):
        backend = MockBackend()
        server = Server(backend, 'tcp://127.0.0.1:31338')
        client = Client('tcp://127.0.0.1:31338')

        try:
            server.start()
            # This test is screwy. If simplejson is installed, it will convert
            # all strings to str objects. Without it, you get unicode objects.
            # We require simplejson so assume it's gonna be str.
            self.assertEqual(client.foo(4, 'blah'), "foo: 4 'blah'")
            self.assertEqual(
                client.bar('hello', 'world', **dict(a=12, b='blah')),
                "bar: ('hello', 'world') {'a': 12, 'b': 'blah'}")

            with self.assertRaisesRegexp(ServerError, 'ValueError: sad'):
                client.failme(42)
        finally:
            server.kill()

    def test_timeout(self):
        client = Client('tcp://127.0.0.1:31339', wait=10)
        with self.assertRaisesRegexp(TimeoutError,
                                     'Timed out after 10 ms waiting'):
            client.foo()

    def _run_server_with_args(self, args, path, url, bind):
        log = TimeRotatingLog(path)
        data.run_clickstream(log)

        drop_existing_tables(create_engine(url))

        sys.argv = args

        killed_event = Event()
        th = Thread(target=server_main, args=(killed_event,))
        orig_interact = code.interact

        try:
            th.start()

            # Give the server time to process all the records before querying
            # it.
            time.sleep(0.5)

            def fake_interact(banner, local):
                client = local['client']
                self.assertEqual(client.count(u'add to cart', site_id=1), 5)
                self.assertEqual(client.count(u'began checkout', site_id=1), 4)
                self.assertEqual(client.count(u'viewed page', site_id=1), 6)
                self.assertEqual(client.count(u'abandoned cart', site_id=1), 1)
                self.assertEqual(
                    client.count(u'abandoned after validation failure',
                                 site_id=1), 0)

            code.interact = fake_interact

            sys.argv = ['manhattan-client', '--connect=%s' % bind]
            client_main()
        finally:
            code.interact = orig_interact
            killed_event.set()
            # Wait for thread to die before returning.
            time.sleep(0.5)

    def test_clientserver_executable(self):
        path = work_path('clientserver-executable')
        log_path = work_path('debug.log')
        url = 'sqlite:////tmp/manhattan-clientserver.db'
        bind = 'tcp://127.0.0.1:5555'

        args = [
            'manhattan-server',
            '--url=%s' % url,
            '--path=%s' % path,
            '--log=%s' % log_path,
            '--bind=%s' % bind,
            '--complex="abandoned cart|add to cart|began checkout"',
            '--complex="abandoned checkout|began checkout|completed checkout"',
            '--complex="abandoned after validation failure|'
            'began checkout,checkout validation failed|completed checkout"',
            '--complex="abandoned after payment failure|'
            'began checkout,payment failed|completed checkout"',
        ]
        self._run_server_with_args(args, path, url, bind)
        self.assertTrue(os.path.exists(log_path))

    def test_configure_logging(self):
        cfg = logging_config(filename=None)
        self.assertNotIn('root_file', cfg)

    def test_clientserver_python_config(self):
        path = data.sampleconfig['input_log_path']
        url = data.sampleconfig['sqlalchemy_url']
        bind = data.sampleconfig['bind']
        log_path = data.sampleconfig['error_log_path']

        args = ['manhattan-server',
                '--config=manhattan.tests.data.sampleconfig']
        self._run_server_with_args(args, path, url, bind)
        self.assertTrue(os.path.exists(log_path))



class TextLog(object):
    """
    A log superclass which provides string handling methods, intended for logs
    which will use line-based text. This class does not implement a usable log
    by itself, and is intended to be subclassed.
    """
    def format(self, elements):
        return b'\t'.join(el.encode('unicode_escape') for el in elements)

    def parse(self, record):
        return [el.decode('unicode_escape') for el in record.split(b'\t')]



import types
import time

from threading import Thread

from manhattan.record import Record, PageRecord, GoalRecord
from manhattan.log.timerotating import TimeRotatingLog

from .base import BaseTest, work_path


def set_fake_name(log, index):
    def fake_name(self, timestamp):
        return '%s.%s' % (self.path, index)
    log.log_name_for = types.MethodType(fake_name, log)


def make_thread_consumer(log_r, process_from=None):
    consumed = []
    last_pointer_container = [None]
    log_r.sleep_delay = 0.001

    def consume(l):
        for rec, ptr in l.process(stay_alive=True, process_from=process_from):
            consumed.append(Record.from_list(rec))
            last_pointer_container[0] = ptr

    consumer = Thread(target=consume, args=(log_r,))
    consumer.start()
    return consumed, consumer, last_pointer_container


class TimeRotatingLogTest(BaseTest):

    def test_basic(self):
        path = work_path('trl-basic')
        log_w = TimeRotatingLog(path)
        log_w.write(PageRecord(url='/foo').to_list())

        log_r = TimeRotatingLog(path)
        records = list(log_r.process(stay_alive=False))
        self.assertEqual(len(records), 1)
        rec = Record.from_list(records[0][0])
        self.assertEqual(rec.url, '/foo')

    def test_multiple_logs(self):
        path = work_path('trl-multi')
        log_w = TimeRotatingLog(path)

        set_fake_name(log_w, '001')
        log_w.write(PageRecord(url='/foo').to_list())

        set_fake_name(log_w, '004')
        log_w.write(PageRecord(url='/bar').to_list())

        log_r = TimeRotatingLog(path)
        records = list(log_r.process(stay_alive=False))
        self.assertEqual(len(records), 2)
        self.assertEqual(Record.from_list(records[0][0]).url, '/foo')
        self.assertEqual(Record.from_list(records[1][0]).url, '/bar')

    def test_stay_alive_single(self):
        path = work_path('trl-stayalive')
        log_r = TimeRotatingLog(path)
        log_r.sleep_delay = 0.001
        consumed, consumer, _ = make_thread_consumer(log_r)

        try:
            self.assertEqual(len(consumed), 0)

            log_w = TimeRotatingLog(path)

            log_w.write(PageRecord(url='/baz').to_list())
            time.sleep(log_r.sleep_delay * 10)

            self.assertEqual(len(consumed), 1)
            self.assertEqual(consumed[0].url, '/baz')

            log_w.write(PageRecord(url='/herp').to_list())
            time.sleep(log_r.sleep_delay * 10)

            self.assertEqual(len(consumed), 2)
            self.assertEqual(consumed[1].url, '/herp')
        finally:
            log_r.killed.set()

    def test_stay_alive_multiple(self):
        path = work_path('trl-stayalive-multi')
        log_r = TimeRotatingLog(path)
        log_r.sleep_delay = 0.001
        consumed, consumer, _ = make_thread_consumer(log_r)

        try:
            self.assertEqual(len(consumed), 0)

            log_w = TimeRotatingLog(path)

            set_fake_name(log_w, '357')
            log_w.write(PageRecord(url='/baz').to_list())
            time.sleep(log_r.sleep_delay * 10)

            self.assertEqual(len(consumed), 1)
            self.assertEqual(consumed[0].url, '/baz')

            set_fake_name(log_w, '358')
            log_w.write(PageRecord(url='/herp').to_list())
            time.sleep(log_r.sleep_delay * 10)

            self.assertEqual(len(consumed), 2)
            self.assertEqual(consumed[1].url, '/herp')
        finally:
            log_r.killed.set()

    def test_stay_alive_nofiles(self):
        log_r = TimeRotatingLog(work_path('trl-stayalive-none'))
        log_r.sleep_delay = 0.001
        consumed, consumer, _ = make_thread_consumer(log_r)
        log_r.killed.set()

    def test_unicode_names(self):
        path = work_path('trl-unicode')
        log_w = TimeRotatingLog(path)
        goal_name = u'Goo\xf6aa\xe1llll!!!'
        rec = GoalRecord(name=goal_name,
                         value='',
                         value_type='',
                         value_format='')
        log_w.write(rec.to_list())

        log_r = TimeRotatingLog(path)
        records = list(log_r.process(stay_alive=False))
        self.assertEqual(len(records), 1)
        rec = Record.from_list(records[0][0])
        self.assertEqual(rec.name, goal_name)

    def test_resume(self):
        path = work_path('trl-resume')
        log_w = TimeRotatingLog(path)

        # Create a thread consumer
        log_r1 = TimeRotatingLog(path)

        consumed, consumer, ptr_container = make_thread_consumer(log_r1)

        try:
            # Write one record
            log_w.write(PageRecord(url='/herp').to_list())
            time.sleep(log_r1.sleep_delay * 10)
            # Check that one record was read.
            self.assertEqual(len(consumed), 1)
            self.assertEqual(consumed[0].url, '/herp')
        finally:
            # Kill the thread
            log_r1.killed.set()
            # Wait for it to die.
            time.sleep(log_r1.sleep_delay * 10)

        last_pointer = ptr_container[0]
        self.assertIsNotNone(last_pointer)

        try:
            # Write one record
            log_w.write(PageRecord(url='/derp').to_list())
            time.sleep(log_r1.sleep_delay * 10)
            # Create a new thread consumer
            log_r2 = TimeRotatingLog(path)
            consumed, consumer, _ = \
                make_thread_consumer(log_r2, process_from=last_pointer)
            time.sleep(log_r2.sleep_delay * 10)
            # Check that the second record was read.
            self.assertEqual(len(consumed), 1)
            self.assertEqual(consumed[0].url, '/derp')
        finally:
            log_r2.killed.set()



from collections import OrderedDict


class DeferredLRUCache(object):
    """
    An LRU cache with deferred (on-request) writing. When values put back into
    the cache, they are flagged as dirty, but are not written to the backend
    until ``.flush()`` is called. When the cache is flushed, all dirty values
    will be written back. Note that ``.flush()`` must be called frequently
    enough that the cache does not self-evict dirty values--this will fail.

    This is NOT thread-safe.
    """
    def __init__(self, get_backend, put_backend, max_size=2000):
        """
        Create a new LRU cache.

        :param get_backend:
            Function which fetches value from the persistent backend.
        :type get_backend:
            Callable with the signature ``func(key)``, returning a ``value``.
        :param put_backend:
            Function which flushes values from the cache to the persistent
            backend.
        :type put_backend:
            Callable with the signature ``func(key, value)``. Return value is
            ignored.
        :param max_size:
            Maximum number of entries to allow in the cache.
        :type max_size:
            int
        """
        self.max_size = max_size
        self.get_backend = get_backend
        self.put_backend = put_backend
        self.entries = OrderedDict()
        self.dirty = set()

    def get(self, key):
        """
        Fetch a value from the cache, reading it from the persistent backend if
        necessary.
        """
        try:
            value = self.entries.pop(key)
        except KeyError:
            value = self.get_backend(key)
        self.entries[key] = value
        self.prune()
        return value

    def put(self, key, value):
        """
        Put a value into the cache, flagging it as dirty to be written back to
        the persistent backend on the next ``flush()`` call.
        """
        if key in self.entries:
            self.entries.pop(key)
        self.entries[key] = value
        self.dirty.add(key)
        self.prune()

    def prune(self):
        """
        Prune the cache object back down to the desired size, if it is larger.
        """
        while len(self.entries) > self.max_size:
            key, value = self.entries.popitem(last=False)
            assert key not in self.dirty

    def flush(self):
        """
        Flush dirty cache entries to the persistent backend.
        """
        to_put = {}
        for key in list(self.dirty):
            to_put[key] = self.entries[key]
        self.put_backend(to_put)
        self.dirty = set()



import logging

from decimal import Decimal

from sqlalchemy import MetaData, create_engine
from sqlalchemy.exc import SAWarning

import warnings
# Filter out pysqlite Decimal loss of precision warning.
warnings.filterwarnings('ignore',
                        '.*pysqlite does \*not\* support Decimal',
                        SAWarning,
                        'sqlalchemy.types')
# Turn unicode bind param warnings into errors.
warnings.filterwarnings('error',
                        r'Unicode type received non-unicode bind',
                        SAWarning,
                        'sqlalchemy.engine.default')

from manhattan.worker import Worker

from manhattan.log.memory import MemoryLog
from manhattan.log.timerotating import TimeRotatingLog

from manhattan.backend import Backend

from . import data
from .base import BaseTest, work_path

logging.basicConfig()
logging.getLogger('sqlalchemy.engine').setLevel(logging.DEBUG)


def drop_existing_tables(engine):
    "Drop all tables, including tables that aren't defined in metadata."
    temp_metadata = MetaData()
    temp_metadata.reflect(bind=engine)
    for table in reversed(temp_metadata.sorted_tables):
        table.drop(bind=engine)


class TestCombinations(BaseTest):

    def _check_backend_queries(self, backend):
        self.assertEqual(backend.count(u'add to cart', site_id=1), 5)
        self.assertEqual(backend.count(u'began checkout', site_id=1), 4)
        self.assertEqual(backend.count(u'viewed page', site_id=1), 6)

        self.assertEqual(backend.count(u'add to cart', site_id=2), 1)
        self.assertEqual(backend.count(u'began checkout', site_id=2), 1)
        self.assertEqual(backend.count(u'completed checkout', site_id=2), 0)
        self.assertEqual(backend.count(u'viewed page', site_id=2), 1)

        num = backend.count(u'completed checkout',
                            variant=(u'red checkout form', u'False'),
                            site_id=1)
        self.assertEqual(num, 3)

        revenue = backend.goal_value(u'completed checkout', site_id=1)
        self.assertEqual(revenue, Decimal('108.19'))

        revenue_nored = backend.goal_value(
            u'completed checkout',
            variant=(u'red checkout form', u'False'), site_id=1)
        self.assertEqual(revenue_nored, Decimal('108.19'))

        noreds = backend.count(variant=(u'red checkout form', u'False'),
                               site_id=1)
        self.assertEqual(noreds, 3)

        margin = backend.goal_value(u'order margin', site_id=1)
        margin = margin.quantize(Decimal('.01'))
        self.assertEqual(margin, Decimal('23.47'))

        margin_per = backend.goal_value(u'margin per session', site_id=1)
        margin_per = margin_per.quantize(Decimal('.01'))
        self.assertEqual(margin_per, Decimal('3.90'))

        margin_per_noreds = backend.goal_value(
            u'margin per session',
            variant=(u'red checkout form', u'False'), site_id=1)
        margin_per_noreds = margin_per_noreds.quantize(Decimal('.01'))
        self.assertEqual(margin_per_noreds, Decimal('7.79'))

        abandoned_carts = backend.count(u'abandoned cart', site_id=1)
        self.assertEqual(abandoned_carts, 1)

        abandoned_checkouts = backend.count(u'abandoned checkout', site_id=1)
        self.assertEqual(abandoned_checkouts, 1)

        abandoned_payment = backend.count(u'abandoned after payment failure',
                                          site_id=1)
        self.assertEqual(abandoned_payment, 1)

        abandoned_validation = backend.count(
            u'abandoned after validation failure', site_id=1)
        self.assertEqual(abandoned_validation, 0)

        self.assertEqual(backend.all_tests(),
                         [('red checkout form', 1602, 7246)])

        self.assertEqual(backend.results('red checkout form',
                                         ['viewed page',
                                          'add to cart',
                                          'began checkout',
                                          'completed checkout'],
                                         site_id=1),
                         {
                             u'True': [1, 0, 1, Decimal('0')],
                             u'False': [3, 0, 3, Decimal('108.19')]
                         })

    def _get_backend(self, reset=False):
        #url = 'mysql://manhattan:quux@localhost/manhattan_test'
        url = 'sqlite:////tmp/manhattan-test.db'
        if reset:
            drop_existing_tables(create_engine(url))
        return Backend(url, flush_every=2, cache_size=5,
                       complex_goals=data.test_complex_goals)

    def test_resume(self):
        path = work_path('resume')

        backend = self._get_backend(reset=True)

        log_w = TimeRotatingLog(path)
        data.run_clickstream(log_w, first=0, last=25)

        log_r1 = TimeRotatingLog(path)
        worker1 = Worker(log_r1, backend)
        worker1.run()

        first_pointer = backend.get_pointer()
        self.assertIsNotNone(first_pointer)

        backend = self._get_backend(reset=False)

        data.run_clickstream(log_w, first=25)
        log_r2 = TimeRotatingLog(path)
        worker2 = Worker(log_r2, backend)
        worker2.run(resume=True)

        second_pointer = backend.get_pointer()
        self.assertIsNotNone(second_pointer)

        self._check_backend_queries(backend)

    def test_basic(self):
        path = work_path('basic')

        backend = self._get_backend(reset=True)

        log_w = TimeRotatingLog(path)
        data.run_clickstream(log_w)

        log_r = TimeRotatingLog(path)
        worker1 = Worker(log_r, backend)
        worker1.run()

        self._check_backend_queries(backend)

    def test_memory_log(self):
        log = MemoryLog()
        data.run_clickstream(log)

        backend = self._get_backend(reset=True)

        worker1 = Worker(log, backend)
        worker1.run(resume=False)

        self._check_backend_queries(backend)



import re
from unittest import TestCase

from webob import Request, Response
from webtest import TestApp, TestRequest

from manhattan.middleware import ManhattanMiddleware
from manhattan.record import Record
from manhattan.log.memory import MemoryLog
from manhattan.util import BadSignature


class SampleApp(object):

    def __call__(self, environ, start_response):
        req = Request(environ)

        if req.path_info.endswith('.txt'):
            s = 'Hello %s' % req.path_info
            resp = Response(s)
            resp.content_type = 'text/plain'
        elif req.path_info.endswith('.iter'):
            resp = Response()
            s = 'Hello %s' % req.path_info.encode('ascii')

            def app_iter(sample):
                for piece in ('<html><body>', sample, '</body>', '</html>'):
                    yield piece
                self.consumed_iter = True
                yield ' '

            self.consumed_iter = False
            resp.content_type = 'text/html'
            resp.app_iter = app_iter(s)
        else:
            s = '<html><body><h1>Hello %s</h1></body></html>' % req.path_info
            resp = Response(s)
            resp.content_type = 'text/html'

        return resp(environ, start_response)


log = MemoryLog()

host_map = {'localhost': 3,
            'example.com': 5}

inner_app = SampleApp()
wrapped_app = ManhattanMiddleware(inner_app, log, 'secret', host_map=host_map)
app = TestApp(wrapped_app)


class TestMiddleware(TestCase):

    def setUp(self):
        app.reset()
        log.purge()

    def process(self):
        records = list(log.process())
        self.assertEqual(len(records), 1)
        record = Record.from_list(records[0][0])
        return record

    def test_request(self):
        resp = app.get('/')

        record = self.process()
        self.assertEqual(record.key, 'page')
        self.assertEqual(record.site_id, '3')
        first_vid = record.vid

        m = re.search('<img (.+)src="(.+)" alt="" />', resp.body)
        pixel_path = m.group(2)
        resp = app.get(pixel_path)
        self.assertEqual(resp.content_type, 'image/gif')

        record = self.process()
        self.assertEqual(record.key, 'pixel')
        self.assertEqual(record.site_id, '3')
        self.assertEqual(first_vid, record.vid)

        resp = app.get('/foo')

        record = self.process()
        self.assertEqual(record.key, 'page')
        self.assertTrue(record.url.endswith('/foo'))
        self.assertEqual(record.site_id, '3')
        self.assertEqual(first_vid, record.vid)

    def test_host_map(self):
        resp = app.get('/hello', extra_environ={'HTTP_HOST': 'example.com'})
        self.assertEqual(resp.content_type, 'text/html')

        record = self.process()
        self.assertEqual(record.key, 'page')
        self.assertTrue(record.url.endswith('/hello'))
        self.assertEqual(record.site_id, '5')

    def test_unknown_host(self):
        resp = app.get('/somepage',
                       extra_environ={'HTTP_HOST':
                                      'supercalifragilicious.com'})
        self.assertEqual(resp.content_type, 'text/html')

        record = self.process()
        self.assertEqual(record.key, 'page')
        self.assertTrue(record.url.endswith('/somepage'))
        self.assertEqual(record.site_id, '0')

    def test_pixel_req(self):
        resp = app.get('/vpixel.gif')
        self.assertEqual(resp.content_type, 'image/gif',
                         'An html response should have a pixel tag.')

    def test_non_html_pixel(self):
        resp = app.get('/non-html-page.txt')
        self.assertNotIn('/vpixel.gif', resp.body,
                         'A non-html response should not have a pixel tag.')

    def test_generator_response(self):
        req = Request.blank('/quux.iter')
        resp = req.get_response(wrapped_app)

        self.assertFalse(inner_app.consumed_iter,
                         'The generator response has been buffered by '
                         'middleware before instead of being returned as an '
                         'iterable.')
        self.assertIn('/vpixel.gif', resp.body)
        self.assertTrue(inner_app.consumed_iter)

    def test_latin1_user_agent(self):
        # Example user agent is latin1-encoded, so should be preserved.
        sample_ua = '\xc0 \xe0 hello'
        app.get('/somepage', extra_environ={'HTTP_USER_AGENT': sample_ua})
        record = self.process()
        self.assertEqual(record.user_agent, sample_ua.decode('latin1'))

    def test_nongetpost_methods_not_processed(self):
        app.put('/somepage')
        app.delete('/somepage')
        app.options('/somepage')
        records = list(log.process())
        self.assertEqual(len(records), 0)

    def test_safari_top_sites_not_counted(self):
        app.get('/blah', headers={'X-Purpose': 'preview'})
        records = list(log.process())
        self.assertEqual(len(records), 0)

    def test_signature_mangled(self):
        app.get('/')
        orig_cookie = app.cookies['manhattan']
        # truncate the last 4 chars, which will blow the sig
        bad_cookie = orig_cookie[:-4]
        bad_request = TestRequest.blank('/', cookies={'manhattan': bad_cookie})
        with self.assertRaises(BadSignature):
            app.request(bad_request)



import json
import os
import time
from threading import Thread
from unittest import TestCase

from manhattan.log.remote import make_redis, server, RemoteLog, RemoteLogServer
from manhattan.log.timerotating import TimeRotatingLog


REDIS_KEY = 'manhattan:testing:log:queue'


class ServerThread(Thread):

    def __init__(self, log_server):
        self.log_server = log_server
        super(ServerThread, self).__init__()

    def run(self):
        self.log_server.run()


class TestRemoteLog(TestCase):

    @classmethod
    def setUpClass(cls):
        cls.text_log = TimeRotatingLog('/tmp/manhattan-tests/remote.log')
        cls.log = RemoteLog(key=REDIS_KEY)
        for f in cls._get_log_files():
            os.remove(f)
        cls.log.db.delete(REDIS_KEY)

    @classmethod
    def _get_log_files(cls):
        files = []
        dirname = os.path.dirname(cls.text_log.path)
        for f in os.listdir(dirname):
            if f.startswith('remote.log'):
                files.append(os.path.join(dirname, f))
        return files

    def test_01_enqueue(self):
        record = ['a', 'b', 'c']
        self.log.write(record)
        item = self.log.db.lpop(REDIS_KEY)
        stored_record = json.loads(item)[0]
        self.assertEqual(stored_record, record)
        self.assertEqual(self.log.db.llen(REDIS_KEY), 0)

    def test_02_consume(self):
        log_server = RemoteLogServer(self.text_log, key=REDIS_KEY)
        log_server_thread = ServerThread(log_server)
        self.log.write(['a', 'b', 'c'])
        self.log.write(['x', 'y', 'z'])
        log_server_thread.start()
        self.log.write(['1', '2', '3'])
        time.sleep(0.5)
        log_file = self._get_log_files()[0]
        with open(log_file) as fp:
            lines = fp.readlines()
            self.assertEqual(len(lines), 3)
        self.assertEqual(self.log.db.llen(REDIS_KEY), 0)
        self.log.send_command('STOP')

    def test_server_script(self):
        """Create and immediately stop server."""
        self.log.send_command('STOP')
        server(['--key', REDIS_KEY])
        self.assertEqual(self.log.db.llen(REDIS_KEY), 0)

    def test_make_redis(self):
        redis = make_redis({'socket_timeout': 1}, socket_timeout=2)
        connection = redis.connection_pool.make_connection()
        self.assertEqual(connection.socket_timeout, 1)
        redis.rpush(REDIS_KEY, 'a')
        self.assertEqual(redis.lpop(REDIS_KEY), 'a')
        self.assertEqual(redis.llen(REDIS_KEY), 0)



import time
import logging

from .record import Record


log = logging.getLogger(__name__)


class Worker(object):

    def __init__(self, log, backend, stats_every=50):
        self.log = log
        self.backend = backend
        self.stats_every = stats_every

        self.last_live_ts = None
        self.last_record_ts = None
        self.last_num_records = None

    def dump_stats(self, num_records, record_ts):
        live_ts = time.time()
        if self.last_live_ts:
            live_elapsed = live_ts - self.last_live_ts
            record_elapsed = record_ts - self.last_record_ts

            record_rate = ((num_records - self.last_num_records) /
                           float(live_elapsed))

            clock_rate = float(record_elapsed) / float(live_elapsed)
            secs_behind = live_ts - record_ts
            clock_eta = secs_behind / clock_rate

            log.info('Processed %d records, %0.1f /sec, %0.1fx realtime, '
                     '%d secs behind, ETA: %0.1f seconds',
                     num_records, record_rate, clock_rate, secs_behind,
                     clock_eta)

        self.last_live_ts = live_ts
        self.last_record_ts = record_ts
        self.last_num_records = num_records

    def run(self, resume=True, **kwargs):
        log.info('Worker started processing.')

        if resume:
            kwargs['process_from'] = self.backend.get_pointer()
            log.info('Resuming from %s', kwargs['process_from'])

        for ii, (vals, pointer) in enumerate(self.log.process(**kwargs)):
            record = Record.from_list(vals)
            self.backend.handle(record, pointer)
            if (ii % self.stats_every) == 0:
                self.dump_stats(ii, int(float(record.timestamp)))

        log.info('Worker finished processing.')



from collections import Counter, defaultdict
from decimal import Decimal
from operator import itemgetter

from manhattan import visitor

from .rollups import AllRollup, LocalDayRollup, LocalWeekRollup, BrowserRollup
from .cache import DeferredLRUCache
from .model import VisitorHistory, Test, Goal

from .persistence.sql import SQLPersistentStore


default_rollups = {
    'all': AllRollup(),
    'pst_day': LocalDayRollup('America/Los_Angeles'),
    'pst_week': LocalWeekRollup('America/Los_Angeles'),
    'browser': BrowserRollup(),
}


class Backend(object):

    def __init__(self, sqlalchemy_url, rollups=None, complex_goals=None,
                 flush_every=500, cache_size=2000):
        self.rollups = rollups or default_rollups
        self.complex_goals = complex_goals or []

        self.store = store = SQLPersistentStore(sqlalchemy_url)

        self.visitors = DeferredLRUCache(get_backend=store.get_visitor_history,
                                         put_backend=store.put_visitor_history,
                                         max_size=cache_size)
        self.tests = DeferredLRUCache(get_backend=store.get_test,
                                      put_backend=store.put_test,
                                      max_size=cache_size)
        self.goals = DeferredLRUCache(get_backend=store.get_goal,
                                      put_backend=store.put_goal,
                                      max_size=cache_size)

        self.pointer = self.store.get_pointer()
        self.records_since_flush = 0
        self.flush_every = flush_every

        self.reset_counters()

    def get_pointer(self):
        return self.pointer

    def reset_counters(self):
        self.inc_conversions = Counter()
        self.inc_values = defaultdict(Decimal)

        self.inc_variant_conversions = Counter()
        self.inc_variant_values = Counter()

        self.inc_impressions = Counter()

    def handle(self, rec, ptr):
        try:
            history = self.visitors.get(rec.vid)
        except KeyError:
            history = VisitorHistory()

        if rec.key == 'pixel':
            history.nonbot = True
            for rec in history.nonbot_queue:
                self.handle_nonbot(rec, history)

        elif history.nonbot:
            self.handle_nonbot(rec, history)

        else:
            history.nonbot_queue.append(rec)
            # Limit nonbot queue to most recent 500 events.
            del history.nonbot_queue[:-500]

        self.visitors.put(rec.vid, history)
        self.pointer = ptr
        self.records_since_flush += 1

        if self.records_since_flush > self.flush_every:
            self.flush()
            self.records_since_flush = 0

    def handle_nonbot(self, rec, history):
        assert rec.key in ('page', 'goal', 'split')
        ts = int(float(rec.timestamp))
        site_id = int(rec.site_id)

        if rec.key == 'page':
            history.ips.add(rec.ip)
            history.user_agents.add(rec.user_agent)
            self.record_conversion(history,
                                   vid=rec.vid,
                                   name=u'viewed page',
                                   timestamp=ts,
                                   site_id=site_id)

        elif rec.key == 'goal':
            self.record_conversion(history,
                                   vid=rec.vid,
                                   name=rec.name,
                                   timestamp=ts,
                                   site_id=site_id,
                                   value=rec.value,
                                   value_type=rec.value_type,
                                   value_format=rec.value_format)

        else:  # split
            self.record_impression(history,
                                   vid=rec.vid,
                                   name=rec.test_name,
                                   selected=rec.selected,
                                   timestamp=ts,
                                   site_id=site_id)

    def record_impression(self, history, vid, name, selected, timestamp,
                          site_id):
        variant = name, selected
        history.variants.add(variant)

        try:
            test = self.tests.get(name)
        except KeyError:
            test = Test()
            test.first_timestamp = timestamp

        test.last_timestamp = timestamp
        test.variants.add(variant)

        self.tests.put(name, test)

        # Record this impression in appropriate time buckets both on the
        # history object and in the current incremental accumulators.
        for rollup_key, rollup in self.rollups.iteritems():
            bucket_id = rollup.get_bucket(timestamp, history)
            key = (name, selected, rollup_key, bucket_id, site_id)
            if key not in history.impression_keys:
                history.impression_keys.add(key)
                self.inc_impressions[key] += 1

    def iter_rollups(self, timestamp, history):
        for rollup_key, rollup in self.rollups.iteritems():
            bucket_id = rollup.get_bucket(timestamp, history)
            yield rollup_key, bucket_id

    def record_complex_goals(self, history, new_name, timestamp, site_id):
        for complex_name, include, exclude in self.complex_goals:
            # If all goals have now been satisfied in the 'include' set,
            # trigger a +1 delta on this complex goal in the current
            # rollups, and track that as a complex goal conversion in this
            # visitor history.
            if (new_name in include) and (history.goals >= include):
                new_keys = []
                for rollup_key, bucket_id in self.iter_rollups(timestamp,
                                                               history):
                    conv_key = (complex_name, rollup_key, bucket_id, site_id)
                    new_keys.append(conv_key)
                    self.inc_conversions[conv_key] += 1
                history.complex_keys[complex_name] = new_keys

            # If we are adding the first goal in the 'exclude' set, trigger
            # a -1 delta for all conversions on that complex goal in the
            # visitory history.
            if history.goals & exclude == set([new_name]):
                for key in history.complex_keys.pop(complex_name, []):
                    self.inc_conversions[key] -= 1

    def record_conversion(self, history, vid, name, timestamp, site_id,
                          value=None, value_type='', value_format=''):
        try:
            goal = self.goals.get(name)
        except KeyError:
            goal = Goal()
            goal.value_type = value_type
            goal.value_format = value_format

        self.goals.put(name, goal)

        if value:
            value = Decimal(value)

        if name not in history.goals:
            history.goals.add(name)
            # If this is a 'new' goal for this visitor, process complex
            # conversion goals.
            self.record_complex_goals(history, name, timestamp, site_id)

        # Record this goal conversion in appropriate time buckets both on the
        # history object and in the current incremental accumulators.
        for rollup_key, bucket_id in self.iter_rollups(timestamp, history):

            conv_key = (name, rollup_key, bucket_id, site_id)
            if conv_key not in history.conversion_keys:
                history.conversion_keys.add(conv_key)
                self.inc_conversions[conv_key] += 1
            if value:
                self.inc_values[conv_key] += value

            for test_name, selected in history.variants:
                vc_key = (name, test_name, selected, rollup_key, bucket_id,
                          site_id)
                if vc_key not in history.variant_conversion_keys:
                    history.variant_conversion_keys.add(vc_key)
                    self.inc_variant_conversions[vc_key] += 1
                if value:
                    self.inc_variant_values[vc_key] += value

    def flush(self):
        self.store.begin()

        self.visitors.flush()
        self.tests.flush()
        self.goals.flush()

        # Add local counter state onto existing persisted counters.
        self.store.increment_conversion_counters(self.inc_conversions,
                                                 self.inc_values)
        self.store.increment_impression_counters(self.inc_impressions)
        self.store.increment_variant_conversion_counters(
            self.inc_variant_conversions, self.inc_variant_values)
        self.reset_counters()

        self.store.update_pointer(self.pointer)
        self.store.commit()

    def count(self, goal=None, variant=None, rollup_key='all', bucket_id=0,
              site_id=None):
        assert goal or variant, "must specify goal or variant"

        if goal and variant:
            test_name, selected = variant
            key = goal, test_name, selected, rollup_key, bucket_id, site_id
            local = self.inc_variant_conversions[key]
            flushed = self.store.count_variant_conversions(*key)[0]
        elif goal:
            key = goal, rollup_key, bucket_id, site_id
            local = self.inc_conversions[key]
            flushed = self.store.count_conversions(*key)[0]
        else:
            # variant
            name, selected = variant
            key = name, selected, rollup_key, bucket_id, site_id
            local = self.inc_impressions[key]
            flushed = self.store.count_impressions(*key)

        return local + flushed

    def goal_value(self, goal, variant=None, rollup_key='all', bucket_id=0,
                   site_id=None):

        goal_obj = self.goals.get(goal)

        if not goal_obj.value_type:
            return self.count(goal, variant, rollup_key=rollup_key,
                              bucket_id=bucket_id, site_id=site_id)

        if variant:
            test_name, selected = variant
            key = goal, test_name, selected, rollup_key, bucket_id, site_id
            local = self.inc_variant_values[key]
            flushed = self.store.count_variant_conversions(*key)[1]
        else:
            key = goal, rollup_key, bucket_id, site_id
            local = self.inc_values[key]
            flushed = self.store.count_conversions(*key)[1]
        value = local + Decimal(str(flushed))

        if goal_obj.value_type == visitor.SUM:
            return value

        elif goal_obj.value_type == visitor.AVERAGE:
            count = self.count(goal, variant, rollup_key=rollup_key,
                               bucket_id=bucket_id, site_id=site_id)
            return value / count if count > 0 else 0

        else:
            # visitor.PER
            count = self.count(u'viewed page', variant,
                               rollup_key=rollup_key,
                               bucket_id=bucket_id, site_id=site_id)
            return value / count if count > 0 else 0

    def all_tests(self):
        # Start with flushed.
        all = self.store.all_tests()
        # Update from unflushed (so that dirty entries overwrite the flushed).
        all.update(self.tests.entries)
        # Sort by last timestamp descending.
        all = [(name, test.first_timestamp, test.last_timestamp)
               for name, test in all.iteritems()]
        all.sort(key=itemgetter(2), reverse=True)
        return all

    def results(self, test_name, goals, site_id=None):
        # Return a dict: keys are populations, values are a list of values for
        # the goals specified.
        test = self.tests.get(test_name)

        ret = {}
        for variant in test.variants:
            values = []
            for goal in goals:
                values.append(self.goal_value(goal, variant, site_id=site_id))
            ret[variant[1]] = values

        return ret



import logging
import logging.config
import sys
import argparse
from threading import Thread

import zmq
from zmq.eventloop import ioloop

import manhattan
from manhattan.worker import Worker
from manhattan.log.timerotating import TimeRotatingLog
from manhattan.backend import Backend

log = logging.getLogger(__name__)


loop = ioloop.IOLoop.instance()
ctx = zmq.Context()
default_bind = 'tcp://127.0.0.1:5555'


class Server(Thread):

    def __init__(self, backend, bind=default_bind):
        Thread.__init__(self)
        self.backend = backend
        self.bind = bind

    def run(self):
        s = ctx.socket(zmq.REP)
        s.bind(self.bind)
        loop.add_handler(s, self.handle_zmq, zmq.POLLIN)
        loop.start()

    def kill(self):
        loop.stop()

    def handle_zmq(self, sock, events):
        try:
            req = sock.recv_json()
            resp = self.handle(req)
            msg = ['ok', resp]
        except Exception as e:
            msg = ['error', '%s: %s' % (e.__class__.__name__, str(e))]
        sock.send_json(msg)

    def handle(self, req):
        log.info('Handling request: %r', req)
        method, args, kwargs = req
        resp = getattr(self.backend, method)(*args, **kwargs)
        log.info('Returning response: %r', resp)
        return resp


def logging_config(verbose=False, filename=None):
    handlers = {
        'console': {
            'class': 'logging.StreamHandler',
            'stream': sys.stderr,
            'formatter': 'generic',
            'level': logging.DEBUG if verbose else logging.WARN,
        },
        'null': {
            'class': 'logging.NullHandler',
        }
    }

    if filename:
        handlers['root_file'] = {
            'class': 'logging.FileHandler',
            'formatter': 'generic',
            'level': 'NOTSET',
            'filename': filename,
        }

    return {
        'version': 1,
        'formatters': {
            'generic': {
                'format':
                "%(asctime)s %(levelname)-5.5s [%(name)s] %(message)s"
            },
        },
        'handlers': handlers,
        'loggers': {
            'manhattan': {
                'propagate': True,
                'level': 'NOTSET',
                'handlers': handlers.keys(),
            },
        },
        'root': {
            'level': 'DEBUG',
            'handlers': ['null']
        }
    }


def parse_names(names):
    return set(unicode(name).strip() for name in names.split(','))


def parse_complex_goals(complex):
    complex = complex or []
    configured = []
    for line in complex:
        complex_name, include_names, exclude_names = line.strip('"').split('|')
        include = parse_names(include_names)
        exclude = parse_names(exclude_names)
        configured.append((unicode(complex_name), include, exclude))
    return configured


def load_args_config(args):
    return dict(verbose=args.verbose,
                error_log_path=args.error_log_path,
                complex_goals=parse_complex_goals(args.complex),
                sqlalchemy_url=args.url,
                input_log_path=args.input_log_path,
                bind=args.bind)


def load_python_config(namespace):
    mod_name, attr_name = namespace.rsplit('.', 1)
    __import__(mod_name)
    mod = sys.modules[mod_name]
    return getattr(mod, attr_name)


def main(killed_event=None):
    p = argparse.ArgumentParser(
        description='Run a Manhattan worker with a TimeRotatingLog.')

    p.add_argument('-v', '--verbose', dest='verbose', action='store_true',
                   default=False, help='Print detailed output')
    p.add_argument('-p', '--path', dest='input_log_path', type=str,
                   help='Input Manhattan log path')
    p.add_argument('--log', dest='error_log_path', type=str,
                   help='Path to error/debug log')
    p.add_argument('-u', '--url', dest='url', type=str,
                   help='SQL backend URL')
    p.add_argument('-c', '--complex', dest='complex', action='append',
                   help='Configure complex goal, like '
                   'name|include a, include b|exclude a')
    p.add_argument('--bind', type=str,
                   default='tcp://127.0.0.1:5555',
                   help='ZeroMQ socket description to bind to')

    p.add_argument('--config', type=str,
                   help='Python namespace to use for configuration')

    args = p.parse_args()

    if args.config:
        config = load_python_config(args.config)
    else:
        config = load_args_config(args)

    logging.config.dictConfig(logging_config(config.pop('verbose'),
                                             config.pop('error_log_path')))

    input_log_path = config.pop('input_log_path')
    bind = config.pop('bind')
    backend = Backend(**config)
    manhattan.server_backend = backend

    mhlog = TimeRotatingLog(input_log_path)
    worker = Worker(mhlog, backend, stats_every=5000)

    server = Server(backend, bind=bind)
    server.start()

    try:
        worker.run(stay_alive=True, killed_event=killed_event)
    finally:
        server.kill()



import logging

from collections import deque

log = logging.getLogger(__name__)


class MemoryLog(object):
    """
    An in-memory log, intended for testing only. This log does not support
    resuming (server crash recovery).
    """
    def __init__(self):
        self.q = deque()

    def write(self, *records):
        log.info('Writing records: %r', records)
        self.q.extend(records)

    def process(self):
        log.info('Swapping out log.')
        to_process = self.q
        log.info('Creating new log.')
        self.q = deque()
        log.info('Playing back old log.')
        for record in to_process:
            log.info('Playing record: %r', record)
            yield record, None

    def purge(self):
        log.info('Purging log.')
        self.q = deque()



"""
This module contains some example ``Rollup`` objects, each implementing the
interface expected by the Manhattan backend for rollup aggregations.
"""

import time
from datetime import datetime, timedelta

import pytz


class LocalRollup(object):

    def __init__(self, tzname):
        self.tz = pytz.timezone(tzname)

    def start_date_for(self, timestamp):
        dt = datetime.utcfromtimestamp(timestamp).replace(tzinfo=pytz.utc)
        dt_local = dt.astimezone(self.tz).replace(tzinfo=None)
        return dt_local.date()


class LocalDayRollup(LocalRollup):

    def get_bucket(self, timestamp, history):
        return time.mktime(self.start_date_for(timestamp).timetuple())


class LocalWeekRollup(LocalRollup):

    def get_bucket(self, timestamp, history):
        day = self.start_date_for(timestamp)
        days_from_sunday = day.isoweekday() % 7
        day -= timedelta(days=days_from_sunday)
        return time.mktime(day.timetuple())


class AllRollup(object):

    def get_bucket(self, timestamp, history):
        return 0


class BrowserRollup(object):

    def browser_from_user_agent(self, user_agent):
        # FIXME This is a pretty naive and less useful implementation.
        if 'Chrome' in user_agent:
            return u'Chrome'
        elif 'Safari' in user_agent:
            return u'Safari'
        elif 'Firefox' in user_agent:
            return u'Firefox'
        elif 'MSIE' in user_agent:
            return u'IE'
        else:
            return u'Unknown'

    def get_bucket(self, timestamp, history):
        return (history.user_agents and
                self.browser_from_user_agent(list(history.user_agents)[0]) or
                u'')



import os
import os.path
import time
import glob
from fcntl import flock, LOCK_EX, LOCK_UN
from threading import Event

from .text import TextLog


class TimeRotatingLog(TextLog):
    """
    A type of log which writes records as individual lines to a series of
    files, with one file per hour of time in which events occur. Files are
    written atomically (using ``fcntl.flock``) and only appended to.
    """
    sleep_delay = 0.5

    def __init__(self, path):
        self.path = path
        self.current_log_name = None
        self.killed = Event()
        self.f = None

    def create_dirs(self):
        dirpath = os.path.dirname(self.path)
        if dirpath and not os.path.exists(dirpath):
            os.makedirs(dirpath)

    def log_name_for(self, ts):
        ts = int(ts)
        start = ts - (ts % 3600)
        return '%s.%s' % (self.path, start)

    def write(self, *records):
        check_log_name = self.log_name_for(time.time())
        if check_log_name != self.current_log_name:
            self.current_log_name = check_log_name
            self.create_dirs()
            if self.f:
                self.f.close()
            self.f = open(self.current_log_name, 'ab')

        data = [self.format(r) for r in records]
        for r in data:
            assert b'\n' not in r, '\\n found in %r' % r
        data.append('')  # to get the final \n
        data = b'\n'.join(data)

        flock(self.f, LOCK_EX)
        self.f.write(data)
        self.f.flush()
        flock(self.f, LOCK_UN)

    def live_iter_glob(self, start_file):
        """
        Yield an infinite iterator of the available log file names. If there
        are no new files to yield, just re-yields the most recent one.
        """
        last_consumed = None
        while True:
            fnames = glob.glob('%s.[0-9]*' % self.path)
            fnames.sort()

            # Crop fnames to start at ``start_file`` if it is supplied.
            if start_file and (start_file in fnames):
                fnames = fnames[fnames.index(start_file):]

            fresh_files = False
            for fn in fnames:
                if (not last_consumed) or (fn > last_consumed):
                    fresh_files = True
                    last_consumed = fn
                    yield fn
            if not fresh_files:
                if fnames:
                    yield fnames[-1]
                elif not self.killed.is_set():
                    time.sleep(self.sleep_delay)
                else:
                    break

    def tail_glob(self, start_file, start_offset):
        """
        Return an iterator over all the matching log files, yielding a line at
        a time. At the end of all available files, poll the last file for new
        lines and look for new files. If a new file is created, abandon the
        previous file and follow that one.
        """
        fnames = self.live_iter_glob(start_file=start_file)
        this_file = next(fnames)
        f = open(this_file, 'rb')

        if start_offset:
            f.seek(start_offset)

        while True:
            start = f.tell()
            line = f.readline()
            if not line:
                next_file = next(fnames)
                if next_file != this_file:
                    this_file = next_file
                    f = open(this_file, 'rb')
                elif not self.killed.is_set():
                    time.sleep(self.sleep_delay)
                    f.seek(start)
                else:
                    break
            else:
                pointer = '%s:%d' % (this_file, f.tell())
                yield line, pointer

    def process(self, process_from=None, stay_alive=False, killed_event=None):
        if not stay_alive:
            self.killed.set()
        if killed_event:
            self.killed = killed_event

        if process_from:
            start_file, start_offset = process_from.rsplit(':', 1)
            start_offset = int(start_offset)
        else:
            start_file = start_offset = None

        for line, pointer in self.tail_glob(start_file=start_file,
                                            start_offset=start_offset):
            yield self.parse(line.strip()), pointer



import argparse
import json

import redis

from manhattan.log.timerotating import TimeRotatingLog


DEFAULT_REDIS_KEY = 'manhattan:log:queue'


def make_redis(kwargs=None, **defaults):
    if kwargs is None:
        kwargs = {}
    for k in defaults:
        if k not in kwargs:
            kwargs[k] = defaults[k]
    return redis.Redis(**kwargs)


class RemoteLog(object):

    """Sends log entries to a remote server."""

    def __init__(self, key=DEFAULT_REDIS_KEY, redis_kwargs=None):
        self.key = key
        self.db = make_redis(redis_kwargs, socket_timeout=1)

    def write(self, *records):
        """Send ``records`` to remote logger."""
        self.db.rpush(self.key, json.dumps(records))

    def send_command(self, command):
        self.db.rpush(self.key, command)


class RemoteLogServer(object):

    """Consumes log entries from a Redis queue and writes them to a log."""

    def __init__(self, log, key=DEFAULT_REDIS_KEY, redis_kwargs=None):
        self.log = log
        self.key = key
        self.db = make_redis(redis_kwargs)
        self.running = False

    def run(self):
        self.running = True
        while self.running:
            records = self.db.blpop(self.key)[1]
            if records == 'STOP':
                self.stop()
            else:
                records = json.loads(records)
                self.log.write(*records)

    def stop(self):
        self.running = False


def server(argv=None):
    parser = argparse.ArgumentParser()
    parser.add_argument('-p', '--path', default='log/manhattan.log')
    parser.add_argument('-k', '--key', default=DEFAULT_REDIS_KEY)
    args = parser.parse_args(argv)
    log_server = RemoteLogServer(TimeRotatingLog(args.path), args.key)
    log_server.run()



import warnings

from sqlalchemy import MetaData, Table, Column, types, create_engine, select
from sqlalchemy.sql import and_
from sqlalchemy.dialects import mysql


# Catch truncated data warnings. These are likely to happen on pickle types
# with a backend field that is not large enough, and they will break
# things horribly.
warnings.filterwarnings('error',
                        'Data truncated for column',
                        Warning,
                        'sqlalchemy.engine.default')


from ..model import Goal, Test


class LargePickleType(types.PickleType):

    def load_dialect_impl(self, dialect):
        if dialect.name == 'mysql':
            return dialect.type_descriptor(mysql.LONGBLOB)  # pragma: nocover
        else:
            return dialect.type_descriptor(types.LargeBinary)


class SQLPersistentStore(object):

    def __init__(self, sqlalchemy_url):
        self.engine = create_engine(sqlalchemy_url, pool_recycle=3600,
                                    strategy='threadlocal')
        self.metadata = MetaData(bind=self.engine)

        self.pointer_table = Table(
            'pointer',
            self.metadata,
            Column('pointer', types.String(255), primary_key=True),
            mysql_engine='InnoDB')

        self.history_table = Table(
            'visitor_histories',
            self.metadata,
            Column('vid', types.String(40), primary_key=True),
            Column('history', LargePickleType, nullable=False),
            mysql_engine='InnoDB')

        self.tests_table = Table(
            'tests',
            self.metadata,
            Column('id', types.Integer, primary_key=True),
            Column('name', types.String(255), nullable=False, index=True),
            Column('first_timestamp', types.Integer, nullable=False),
            Column('last_timestamp', types.Integer, nullable=False),
            Column('variants', LargePickleType, nullable=False),
            mysql_engine='InnoDB')

        self.goal_table = Table(
            'goals',
            self.metadata,
            Column('id', types.Integer, primary_key=True),
            Column('name', types.String(255), nullable=False, index=True),
            Column('value_type', types.CHAR(1), nullable=False, default=''),
            Column('value_format', types.CHAR(1), nullable=False, default=''),
            mysql_engine='InnoDB')

        self.conversion_counts_table = Table(
            'conversion_counts',
            self.metadata,
            Column('name', types.String(255), primary_key=True),
            Column('rollup_key', types.String(255), primary_key=True),
            Column('bucket_id', types.String(255), primary_key=True),
            Column('site_id', types.Integer, primary_key=True),
            Column('count', types.Integer, nullable=False, default=0),
            Column('value', types.Float, nullable=False, default=0),
            mysql_engine='InnoDB')

        self.impression_counts_table = Table(
            'impression_counts',
            self.metadata,
            Column('name', types.String(255), primary_key=True),
            Column('selected', types.String(255), primary_key=True),
            Column('rollup_key', types.String(255), primary_key=True),
            Column('bucket_id', types.String(255), primary_key=True),
            Column('site_id', types.Integer, primary_key=True),
            Column('count', types.Integer, nullable=False, default=0),
            mysql_engine='InnoDB')

        self.variant_conversion_counts_table = Table(
            'variant_conversion_counts',
            self.metadata,
            Column('goal_name', types.String(100), primary_key=True),
            Column('test_name', types.String(100), primary_key=True),
            Column('selected', types.String(100), primary_key=True),
            Column('rollup_key', types.String(100), primary_key=True),
            Column('bucket_id', types.String(100), primary_key=True),
            Column('site_id', types.Integer, primary_key=True),
            Column('count', types.Integer, nullable=False, default=0),
            Column('value', types.Float, nullable=False, default=0),
            mysql_engine='InnoDB')

        self.metadata.create_all()

    def update_pointer(self, ptr):
        if ptr is None:
            return
        q = self.pointer_table.update().values(pointer=ptr)
        r = q.execute()
        if r.rowcount == 0:
            q = self.pointer_table.insert().values(pointer=ptr)
            q.execute()

    def get_pointer(self):
        return select([self.pointer_table.c.pointer]).scalar()

    def begin(self):
        return self.engine.begin()

    def commit(self):
        return self.engine.commit()

    def criteria_from_dict(self, table, key_dict):
        criteria = []
        for col, val in key_dict.iteritems():
            criteria.append(getattr(table.c, col) == val)
        if len(criteria) > 1:
            return and_(*criteria)
        else:
            return criteria[0]

    def put_kv(self, table, key_dict, value_dict, increment=False):
        whereclause = self.criteria_from_dict(table, key_dict)

        if increment:
            update_dict = {}
            for col, val in value_dict.iteritems():
                update_dict[col] = getattr(table.c, col) + val
        else:
            update_dict = value_dict

        q = table.update().values(**update_dict).where(whereclause)
        r = q.execute()
        if r.rowcount == 0:
            value_dict.update(key_dict)
            q = table.insert().values(**value_dict)
            q.execute()

    def put_visitor_history(self, histories):
        for vid, history in histories.iteritems():
            self.put_kv(self.history_table, {'vid': vid}, {'history': history})

    def put_test(self, tests):
        for name, test in tests.iteritems():
            self.put_kv(self.tests_table,
                        {'name': name},
                        {'first_timestamp': test.first_timestamp,
                         'last_timestamp': test.last_timestamp,
                         'variants': test.variants})

    def put_goal(self, goals):
        for name, goal in goals.iteritems():
            self.put_kv(self.goal_table,
                        {'name': name},
                        {'value_type': goal.value_type,
                         'value_format': goal.value_format})

    def increment_conversion_counters(self, inc_conversions, inc_values):
        """
        Given a map of (goal name, rollup key, bucket start) tuples to
        tuples of (integer counts, Decimal values), adjust the state of the
        counts in the SQL database.
        """
        for key in set(inc_conversions) | set(inc_values):
            delta = inc_conversions.get(key, 0)
            value = inc_values.get(key, 0)
            name, rollup_key, bucket_id, site_id = key

            self.put_kv(self.conversion_counts_table,
                        {'name': name,
                         'rollup_key': rollup_key,
                         'bucket_id': bucket_id,
                         'site_id': site_id},
                        {'count': delta,
                         'value': value},
                        increment=True)

    def increment_impression_counters(self, inc_impressions):
        for key, delta in inc_impressions.iteritems():
            name, selected, rollup_key, bucket_id, site_id = key

            self.put_kv(self.impression_counts_table,
                        {'name': name,
                         'selected': selected,
                         'rollup_key': rollup_key,
                         'bucket_id': bucket_id,
                         'site_id': site_id},
                        {'count': delta},
                        increment=True)

    def increment_variant_conversion_counters(self, inc_variant_conversions,
                                              inc_variant_values):
        for key in set(inc_variant_conversions) | set(inc_variant_values):
            delta = inc_variant_conversions.get(key, 0)
            value = inc_variant_values.get(key, 0)
            goal_name, test_name, selected, \
                rollup_key, bucket_id, site_id = key

            self.put_kv(self.variant_conversion_counts_table,
                        {'goal_name': goal_name,
                         'test_name': test_name,
                         'selected': selected,
                         'rollup_key': rollup_key,
                         'bucket_id': bucket_id,
                         'site_id': site_id},
                        {'count': delta,
                         'value': value},
                        increment=True)

    def get_kv(self, table, get_cols, key_dict, default=None):
        to_select = [getattr(table.c, col) for col in get_cols]
        whereclause = self.criteria_from_dict(table, key_dict)

        r = select(to_select).where(whereclause).execute().first()
        if r:
            return r
        else:
            if default:
                return default
            else:
                raise KeyError

    def get_visitor_history(self, vid):
        r = self.get_kv(self.history_table,
                        ['history'],
                        {'vid': vid})
        return r[0]

    def get_test(self, name):
        r = self.get_kv(self.tests_table,
                        ['first_timestamp', 'last_timestamp', 'variants'],
                        {'name': name})
        first, last, variants = r
        return Test(first_timestamp=first,
                    last_timestamp=last,
                    variants=variants)

    def get_goal(self, name):
        r = self.get_kv(self.goal_table,
                        ['value_type', 'value_format'],
                        {'name': name})
        value_type, value_format = r
        return Goal(value_type=value_type, value_format=value_format)

    def count_conversions(self, name, rollup_key, bucket_id, site_id):
        return self.get_kv(self.conversion_counts_table,
                           ['count', 'value'],
                           {'name': name,
                            'rollup_key': rollup_key,
                            'bucket_id': bucket_id,
                            'site_id': site_id},
                           default=(0, 0))

    def count_impressions(self, name, selected, rollup_key, bucket_id,
                          site_id):
        r = self.get_kv(self.impression_counts_table,
                        ['count'],
                        {'name': name,
                         'selected': selected,
                         'rollup_key': rollup_key,
                         'bucket_id': bucket_id,
                         'site_id': site_id},
                        default=(0,))
        return r[0]

    def count_variant_conversions(self, goal_name, test_name, selected,
                                  rollup_key, bucket_id, site_id):
        return self.get_kv(self.variant_conversion_counts_table,
                           ['count', 'value'],
                           {'goal_name': goal_name,
                            'test_name': test_name,
                            'selected': selected,
                            'rollup_key': rollup_key,
                            'bucket_id': bucket_id,
                            'site_id': site_id},
                           default=(0, 0))

    def all_tests(self):
        t = self.tests_table
        r = select([t.c.name,
                    t.c.first_timestamp,
                    t.c.last_timestamp,
                    t.c.variants]).execute()
        ret = {}
        for name, first, last, variants in r:
            ret[name] = Test(first_timestamp=first,
                             last_timestamp=last,
                             variants=variants)
        return ret



from setuptools import setup, find_packages


setup(name="manhattan",
      version='0.2',
      description='Robust Server-Side Analytics',
      long_description='',
      classifiers=[
          'Development Status :: 5 - Production/Stable',
          'Programming Language :: Python :: 2.7',
          'Topic :: Internet :: WWW/HTTP :: WSGI :: Middleware',
          'Topic :: Internet :: WWW/HTTP :: Site Management',
      ],
      keywords='',
      url='http://github.com/cartlogic/manhattan',
      author='Scott Torborg',
      author_email='scott@cartlogic.com',
      install_requires=[
          'sqlalchemy>=0.7',
          'webob',
          'redis>=2.7.2',
          'pytz',
          'pyzmq',
          'simplejson',
      ],
      license='MIT',
      packages=find_packages(),
      entry_points=dict(
          console_scripts=[
              'manhattan-server=manhattan.server:main',
              'manhattan-client=manhattan.client:main',
              'manhattan-log-server=manhattan.log.remote:server',
          ]
      ),
      test_suite='nose.collector',
      tests_require=['nose'],
      zip_safe=False)



"""
Various utility functions used by Manhattan. It is not expected that these will
be used externally.
"""

import os
import random
import bisect
import hmac
import hashlib
import binascii


"""
A 1 pixel transparent GIF as a bytestring. For use as a tracking "beacon" in an
HTML document.
"""
transparent_pixel = (
    'GIF89a'
    '\x01\x00\x01\x00\x80\x00\x00\xff\xff\xff\x00\x00\x00!\xf9\x04'
    '\x00\x00\x00\x00\x00,\x00\x00\x00\x00\x01\x00\x01\x00\x00\x02'
    '\x02D\x01\x00;')


def pixel_tag(path):
    """
    Build the HTML surrounding a given image path that will be injected as a
    tracking pixel.

    :param path:
        URL path to the tracking pixel.
    :type request:
        string
    :returns:
        HTML markup for image tag.
    :rtype:
        string
    """
    return ('<img style="height:0;width:0;position:absolute;" '
            'src="%s" alt="" />' % path)


def nonrandom_choice(seed, seq):
    """
    Pick an element from the specified sequence ``seq`` based on the value of
    the specified string ``seed``. Guaranteed to be deterministic, tries to be
    uniform.

    :param seed:
        Any string specifier to control the element.
    :type seed:
        string
    :param seq:
        Python sequence to pick from.
    :type seq:
        sequence object
    :returns:
        An element from ``seq``.
    """
    return random.Random(seed).choice(seq)


def nonrandom(seed, n):
    """
    Return a deterministic but pseudo-random number between 0 and ``n``, based
    on the value of ``seed``. Guaranteed to be deterministic, tries to be
    uniform.

    :param seed:
        Any string specifier to control the output.
    :type seed:
        string
    :param n:
        Range of output value.
    :type n:
        float
    :returns:
        Float between 0 and ``n``.
    :rtype:
        float
    """
    return random.Random(seed).random() * n


def choose_population(seed, populations=None):
    """
    Randomly pick an element from populations according to type.

    :param seed:
        Any string specifier to control the output.
    :type seed:
        string
    :param populations:
        If not specified, perform a straight AB test between True and False. If
        specified as a sequence, pick uniformly from the sequence. If specified
        as a dict, use the keys as population names and the values as the
        weight that each population should receive, and distribute between them
        according to weight.
    :type populations:
        None, sequence, or dict
    :returns:
        Selected population
    """
    # Uniform distribution between True, False
    if populations is None:
        return nonrandom_choice(seed, (True, False))

    # Uniform distribution over populations
    if isinstance(populations, list):
        return nonrandom_choice(seed, populations)

    # Weighted distribution over populations
    if isinstance(populations, dict):
        pop_name = []
        pop_mass = []
        running_mass = 0
        for name, mass in populations.iteritems():
            if mass > 0:
                pop_name.append(name)
                pop_mass.append(running_mass)
                running_mass += mass

        if running_mass == 0:
            raise ValueError("Need at least one option with probability > 0")

        r = nonrandom(seed, running_mass)
        i = bisect.bisect(pop_mass, r) - 1
        return pop_name[i]

    raise ValueError("Invalid population description")


def decode_http_header(raw):
    """
    Decode a raw HTTP header into a unicode string. RFC 2616 specifies that
    they should be latin1-encoded (a.k.a. iso-8859-1). If the passed-in value
    is None, return an empty unicode string.

    :param raw:
        Raw HTTP header string.
    :type raw:
        string (non-unicode)
    :returns:
        Decoded HTTP header.
    :rtype:
        unicode string
    """
    if raw:
        return raw.decode('iso-8859-1', 'replace')
    else:
        return u''


def decode_url(raw):
    """
    Decode a URL into a unicode string. Expected to be UTF-8.

    :param raw:
        Raw URL string.
    :type raw:
        string (non-unicode)
    :returns:
        Decode URL.
    :rtype:
        unicode string
    """
    return raw.decode('utf-8')


def constant_time_compare(a, b):
    "Compare two strings with constant time. Used to prevent timing attacks."
    if len(a) != len(b):
        return False
    result = 0
    for x, y in zip(a, b):
        result |= ord(x) ^ ord(y)
    return result == 0


def nonce():
    """
    Build a cryptographically random nonce.

    :returns:
        Hex string, with 20 bytes (40 hex chars).
    :rtype:
        string
    """
    return binascii.hexlify(os.urandom(20))


class SignerError(Exception):
    pass


class BadData(SignerError):
    pass


class BadSignature(SignerError):
    pass


class Signer(object):
    def __init__(self, secret):
        self.key = hashlib.sha1('manhattan.signer.' + secret).digest()
        self.sep = '.'

    def get_signature(self, value):
        "Compute the signature for the given value."
        mac = hmac.new(self.key, msg=value, digestmod=hashlib.sha1)
        return binascii.hexlify(mac.digest())

    def sign(self, value):
        return "%s%s%s" % (value, self.sep, self.get_signature(value))

    def unsign(self, signed_value):
        if self.sep not in signed_value:
            raise BadData('No separator %r found in cookie: %s' %
                          (self.sep, signed_value))

        signed_value = signed_value.lower()
        value, sig = signed_value.rsplit(self.sep, 1)
        expected = self.get_signature(value)

        if constant_time_compare(sig, expected):
            return value

        s = ('Signature for cookie %r does not match: expected %r, '
             'got %s' % (signed_value, expected, sig))
        raise BadSignature(s)



import argparse
import zmq
import code

default_bind = 'tcp://127.0.0.1:5555'


ctx = zmq.Context()


class ServerError(Exception):
    pass


class TimeoutError(Exception):
    pass


class Client(object):

    def __init__(self, connect=default_bind, wait=3000):
        self.sock = ctx.socket(zmq.REQ)
        self.sock.setsockopt(zmq.LINGER, 0)
        self.sock.connect(connect)

        self.poller = zmq.Poller()
        self.poller.register(self.sock, zmq.POLLIN)

        self.wait = wait

    def __getattr__(self, name):
        def rpc_method(*args, **kwargs):
            req = [name, args, kwargs]
            self.sock.send_json(req)

            if self.poller.poll(self.wait):
                status, resp = self.sock.recv_json()
                if status == 'ok':
                    return resp
                else:
                    raise ServerError(resp)
            else:
                raise TimeoutError('Timed out after %d ms waiting for reply' %
                                   self.wait)
        return rpc_method


def main():
    p = argparse.ArgumentParser(description='Run a Manhattan client.')
    p.add_argument('--connect', type=str,
                   default=default_bind,
                   help='ZeroMQ socket description to connect to')
    args = p.parse_args()

    client = Client(args.connect)
    code.interact("The 'client' object is available for queries.",
                  local=dict(client=client))
